#include "mex.h"
#include <math.h>
#include <iostream>
#include <sys/time.h>
#include <vector>
#include <numeric>
#include <inttypes.h>

using namespace std;
#define printf mexPrintf

#include "invertMatrix.h"

//constant time box-filter implementation
void boxfilter3dref(double* vol1,double* temp1,double* temp2,double* ones,int hw,int m,int n,int o){
    const int RANGE=hw;
    const int sz=m*n*o;
    for(int k=0;k<o;k++){
        for(int j=0;j<n;j++){
            double sumA=0;
            const double* inRow=vol1+j*m+k*m*n;
            double* outRow=temp2+j*m+k*m*n;
            
            int x = -RANGE;
            for (; x<=RANGE; x++) {
                sumA += inRow[x+RANGE]; ;
                if (x >= 0) outRow[x] = sumA;
            }
            for (; x+RANGE<m; x++) {
                sumA -= inRow[x-RANGE-1];
                sumA += inRow[x+RANGE];
                outRow[x] = sumA;// * k;
            }
            for (; x < m; x++) {
                sumA -= inRow[x - RANGE - 1]; //sum1 -= 1.0f;
                outRow[x] = sumA;///sum1;
            }
        }
    }
    for(int k=0;k<o;k++){
        for(int i=0;i<m;i++){
            double sumA=0;
            int x = -RANGE;
            for (; x<=RANGE; x++) {
                sumA += temp2[i+(x+RANGE)*m+k*m*n];
                if (x >= 0)
                    temp1[i+x*m+k*m*n]=sumA; //output=>temp1 !!!!
            }
            for (; x+RANGE<n; x++) {
                sumA -= temp2[i+(x-RANGE-1)*m+k*m*n];
                sumA += temp2[i+(x+RANGE)*m+k*m*n];
                temp1[i+x*m+k*m*n]= sumA;// * k;
            }
            for (; x < n; x++) {
                sumA -= temp2[i+(x-RANGE-1)*m+k*m*n];
                temp1[i+x*m+k*m*n] = sumA;///sum1;
            }
        }
    }
    for(int j=0;j<n;j++){
        for(int i=0;i<m;i++){
            double sumA=0;
            int x = -RANGE;
            for (; x<=RANGE; x++) {
                sumA += temp1[i+j*m+(x+RANGE)*m*n];
                if (x >= 0)
                    vol1[i+j*m+x*m*n]=sumA*ones[i+j*m+x*m*n];
            }
            for (; x+RANGE<o; x++) {
                sumA -= temp1[i+j*m+(x-RANGE-1)*m*n];
                sumA += temp1[i+j*m+(x+RANGE)*m*n];
                vol1[i+j*m+x*m*n]= sumA*ones[i+j*m+x*m*n];// * k;
            }
            for (; x < o; x++) {
                sumA -= temp1[i+j*m+(x-RANGE-1)*m*n];
                vol1[i+j*m+x*m*n] = sumA*ones[i+j*m+x*m*n];///sum1;
            }
        }
    }
}

void ccaFilter(double* similarity,double* image1,double* image2,int m,int n,int o,int ch1,int ch2,int r,double eps){
    timeval time1,time2;
    int sz=m*n*o;
    gettimeofday(&time1, NULL);

    double* temp1=new double[sz];
    double* temp2=new double[sz];
    double* ones0=new double[sz];
    
    //set-up boxfilter (number of voxels in every box)
    double* ones=new double[sz];
    for(int i=0;i<sz;i++){
        ones[i]=1.0;
        ones0[i]=1.0;
    }
    boxfilter3dref(ones,temp1,temp2,ones0,r,m,n,o);
    delete ones0;
    for(int i=0;i<sz;i++){
        ones[i]=1.0/(double)ones[i];
    }
    
    //calculate local means and variances of both images using boxfilters
    double* mean1=new double[sz*ch1];
    double* mean2=new double[sz*ch2];
    double* var1=new double[sz*ch1*ch1];
    double* var2=new double[sz*ch2*ch2];
    double* cov=new double[sz*ch1*ch2];
    
    //image1
    for(int c=0;c<ch1;c++){
        for(int i=0;i<sz;i++){
            mean1[i+c*sz]=image1[i+c*sz];
        }
    }
    for(int c1=0;c1<ch1;c1++){
        for(int c2=0;c2<ch1;c2++){
            for(int i=0;i<sz;i++){
                var1[i+(c1+c2*ch1)*sz]=image1[i+c1*sz]*image1[i+c2*sz];
            }
        }
    }
    //image2
    for(int c=0;c<ch2;c++){
        for(int i=0;i<sz;i++){
            mean2[i+c*sz]=image2[i+c*sz];
        }
    }
    for(int c1=0;c1<ch2;c1++){
        for(int c2=0;c2<ch2;c2++){
            for(int i=0;i<sz;i++){
                var2[i+(c1+c2*ch2)*sz]=image2[i+c1*sz]*image2[i+c2*sz];
            }
        }
    }
    //cross-images
    for(int c1=0;c1<ch1;c1++){
        for(int c2=0;c2<ch2;c2++){
            for(int i=0;i<sz;i++){
                cov[i+(c1+c2*ch1)*sz]=image1[i+c1*sz]*image2[i+c2*sz];
            }
        }
    }
     
    //boxfilter
    for(int c=0;c<ch1;c++){
        boxfilter3dref(mean1+c*sz,temp1,temp2,ones,r,m,n,o);
    }
    for(int c=0;c<ch2;c++){
        boxfilter3dref(mean2+c*sz,temp1,temp2,ones,r,m,n,o);
    }
    for(int c=0;c<ch1*ch1;c++){
        boxfilter3dref(var1+c*sz,temp1,temp2,ones,r,m,n,o);
    }
    for(int c=0;c<ch2*ch2;c++){
        boxfilter3dref(var2+c*sz,temp1,temp2,ones,r,m,n,o);
    }
    for(int c=0;c<ch1*ch2;c++){
        boxfilter3dref(cov+c*sz,temp1,temp2,ones,r,m,n,o);
    }
    
    //subtract product of means from (co-)variance
    //image1
    for(int c1=0;c1<ch1;c1++){
        for(int c2=0;c2<ch1;c2++){
            for(int i=0;i<sz;i++){
                var1[i+(c1+c2*ch1)*sz]-=mean1[i+c1*sz]*mean1[i+c2*sz];
            }
        }
    }
    //image2
    for(int c1=0;c1<ch2;c1++){
        for(int c2=0;c2<ch2;c2++){
            for(int i=0;i<sz;i++){
                var2[i+(c1+c2*ch2)*sz]-=mean2[i+c1*sz]*mean2[i+c2*sz];
            }
        }
    }
    //cross-images
    for(int c1=0;c1<ch1;c1++){
        for(int c2=0;c2<ch2;c2++){
            for(int i=0;i<sz;i++){
                cov[i+(c1+c2*ch1)*sz]-=mean1[i+c1*sz]*mean2[i+c2*sz];
            }
        }
    }
    gettimeofday(&time2, NULL);
	double timeP1=time2.tv_sec+time2.tv_usec/1e6-(time1.tv_sec+time1.tv_usec/1e6);
    

    //CCA-calculation voxelwise
    int dmin=min(ch1,ch2);
   
    gettimeofday(&time1, NULL);

    double* Cxy=new double[ch1*ch2];
    double* Cyx=new double[ch2*ch1];
    double* Dx=new double[ch1*ch2];
    double* Dy=new double[ch2*ch1];
    double* D=new double[ch1]; //only diagonal
    
    double* Cxx=new double[ch1*ch1];
    double* iCxx=new double[ch1*ch1];
    double* Cyy=new double[ch2*ch2];
    double* iCyy=new double[ch2*ch2];
    
    double* QRx=new double[2*ch1*ch1+ch1]; //memory for QR-decompositon
    double* QRy=new double[2*ch2*ch2+ch2];

    for(int i=0;i<sz;i++){
        //cross-covariances
        for(int c1=0;c1<ch1;c1++){
            for(int c2=0;c2<ch2;c2++){
                Cxy[c1+c2*ch1]=cov[i+(c1+c2*ch1)*sz];
                Cyx[c2+c1*ch2]=cov[i+(c1+c2*ch1)*sz];
            }
        }
        //inverse var1 (add eps to diagonal)
        for(int c1=0;c1<ch1;c1++){
            for(int c2=0;c2<ch1;c2++){
                Cxx[c1+c2*ch1]=var1[i+(c1+c2*ch1)*sz];
            }
        }
        for(int c=0;c<ch1;c++){
            Cxx[c+c*ch1]+=eps;
        }
        if(ch1==2)
            inv2D(iCxx,Cxx);
        if(ch1==1)
            iCxx[0]=1.0/Cxx[0];
        if(ch1>2)
            qrInv(QRx,iCxx,Cxx,ch1);
        //inverse var2 (add eps to diagonal)
        for(int c1=0;c1<ch2;c1++){
            for(int c2=0;c2<ch2;c2++){
                Cyy[c1+c2*ch2]=var2[i+(c1+c2*ch2)*sz];
            }
        }
        for(int c=0;c<ch2;c++){
            Cyy[c+c*ch2]+=eps;
        }
        if(ch2==2)
            inv2D(iCyy,Cyy);
        if(ch2==1)
            iCyy[0]=1.0/Cyy[0];
        if(ch2>2)
            qrInv(QRy,iCyy,Cyy,ch2);

        //matrix multiplication
        matmult(iCxx,Cxy,Dx,ch1,ch2);
        matmult(iCyy,Cyx,Dy,ch2,ch1);
        
        for(int c=0;c<ch1;c++){
            D[c]=0.0;
            for(int c1=0;c1<ch2;c1++){
                D[c]+=Dx[c+c1*ch1]*Dy[c1+c*ch2];
            }
        }
        
        //trace of correlation matrix
        similarity[i]=1.0;;
        for(int c=0;c<ch1;c++){
            similarity[i]-=D[c]/(double)dmin;
        }
        similarity[i]=min(max(similarity[i],0.0),1.0); //correct for rounding errors
    }
    
    gettimeofday(&time2, NULL);
	double timeP2=time2.tv_sec+time2.tv_usec/1e6-(time1.tv_sec+time1.tv_usec/1e6);
    
    printf("Computation time boxfilter: %4.2f msecs, and CCA: %4.2f msecs\n",1000.0*timeP1,1000.0*timeP2);
    
    delete mean1; delete mean2;
    delete var1; delete var2; delete cov;
    delete temp1; delete ones;

}

/* the gateway function */
void mexFunction( int nlhs, mxArray *plhs[],int nrhs, const mxArray *prhs[]){
	timeval time1,time2;
    
    if(nrhs<4){
        printf("Four inputs required, two double volumes, radius and eps.\n");
        mexErrMsgTxt("Example usage: similarity=LCCA_3d(double(img1),double(img2),5,0.001);");
    }
    //multi channel 2D images / important double precision
    double* image1=(double*)mxGetData(prhs[0]);
    double* image2=(double*)mxGetData(prhs[1]);

    if(not(mxIsDouble(prhs[0]))|not(mxIsDouble(prhs[0]))){
        mexErrMsgTxt("Both input arrays must be double.\n");
    }
    //radius (half-width) of patch
    int r=(int)mxGetScalar(prhs[2]);
    
    //small constant to avoid division by zero (for zero variance patches)
    double eps=(double)mxGetScalar(prhs[3]);

    const mwSize* dims1=mxGetDimensions(prhs[0]);
    int m=dims1[0]; int n=dims1[1]; int o=dims1[2]; int ch1=dims1[3];
    if(mxGetNumberOfDimensions(prhs[0])==3){
        ch1=1;
    }
    const mwSize* dims2=mxGetDimensions(prhs[1]);
    int ch2=dims2[3]; int m2=dims2[0]; int n2=dims2[1]; int o2=dims2[2];
    if(mxGetNumberOfDimensions(prhs[1])==3){
        ch2=1;
    }
    if((m2!=m)|(n2!=n)|(o2!=o)){
        mexErrMsgTxt("Volumes must have same width, height and depth!\n");
    }
    

    int dimsout[]={m,n,o};
    plhs[0]=mxCreateNumericArray(3,dimsout,mxDOUBLE_CLASS,mxREAL);
    double* similarity=(double*)mxGetData(plhs[0]);

    int dmin=min(ch1,ch2);
    printf("dmin: %d, ch1=%d, ch2=%d\n",dmin,ch1,ch2);
    
    
    ccaFilter(similarity,image1,image2,m,n,o,ch1,ch2,r,eps);

    
}
